//************************************************************
//
// BPNGLoggerDetector.cc
//
/** @file BPNGLoggerDetector.cc
 *
 * @brief Description
 *
 * Description
 */
//************************************************************

// sys
#include <iomanip>
#include <iostream>
#include <sstream>
#include <string.h>
#include <ctime>
#include <map>

// client
#include "BPNGLoggerDetector.hh"

using namespace std;


//************************************************************
//
// BPNGLoggerDetector
//
//************************************************************

int BPNGLoggerDetector::onProgressDataDownload(int percentCompleted)
{
    if (percentCompleted != oldpercent)
    {
        cout << "LISTENER: Data download: " << setw(3) << setfill(' ') << percentCompleted << "%" << endl;
        oldpercent = percentCompleted;
    }
    return 1;
}

int BPNGLoggerDetector::onProgressDataDownload(int percentCompleted, uint64_t downloadedSize, uint64_t totalSize)
{
    if (percentCompleted != oldpercent)
    {
        cout << "LISTENER: Data download: " << setw(3) << setfill(' ') << percentCompleted << "%"
             << ", total size: " << totalSize << "B, downloaded size: " << downloadedSize << "B" << endl;
        oldpercent = percentCompleted;
    }
    return 1;
}

int BPNGLoggerDetector::onProgressConversion(int percentCompleted, const char *status)
{
    if (percentCompleted != oldpercent)
    {
        cout << "LISTENER: Conversion: " << setw(3) << setfill(' ') << percentCompleted << "%  " << status << endl;
        oldpercent = percentCompleted;
    }
    return 1;
}

int BPNGLoggerDetector::onProgressDeletion(int percentCompleted)
{
    return 1;
}

void BPNGLoggerDetector::onBPNGDeviceDetected(OnlineLoggerInfo *info)
{
    cout << "LISTENER: Found logger: " << info->ip << endl
         << "          Mainboard: " << info->mbnr << endl
         << "          Name: " << info->name << endl;

    for (const auto &i : loggerList)
        if (string(i.getMbnr()) == info->mbnr)
	    return;

    loggerList.emplace_back(*info);
}


void BPNGLoggerDetector::onStatusMessage(const char *statusMsg)
{
    cout << "LISTENER: " << statusMsg << endl;
}


void BPNGLoggerDetector::onBPNGDeviceDisappeared(OnlineLoggerInfo *info)
{
    auto il = loggerList.begin();
    while (il != loggerList.end())
    {
        if (string(il->getMbnr()) == info->mbnr)
        {
            loggerList.erase(il);
            break;
        }
    }
}


void BPNGLoggerDetector::onBPNGDeviceStateChange(OnlineLoggerInfo *info)
{
    for (auto &i : loggerList)
    {
        if (string(i.getMbnr()) == info->mbnr)
        {
            i = *info;
            break;
        }
    }
}


OnlineLoggerInfo
BPNGLoggerDetector::getLoggerInfoForIP(const char *ip) const
{
    //cout << __func__ << ": ip=\"" << ip << '"' << endl;

    for (const auto &i : loggerList)
    {
        //cout << __func__ << ": i.ip=\"" << i.ip << '"' << endl;
        if (string(i.getIP()) == ip)
            return i.getInfo();
    }

    cerr << "No logger with IP address " << ip << " found!" << endl;

    return createEmptyOnlineLoggerInfo(); // return empty logger info
}


// Returns a vector of BP2Devices, search the network for searchTimeOut seconds
vector<OnlineLoggerInfoWrapper>
BPNGLoggerDetector::getLoggerList(unsigned searchTimeOut)
{
    loggerList.clear();

    IBPNGClient *client = getBPNGClient();

    client->addListener(this);

    time_t start = time(nullptr);

    if (searchTimeOut == 0)
    {
        // scan only once
        client->scanNetworkForLogger();
    }
    else
    {
        while (time(nullptr) - start < searchTimeOut)
        {
            client->scanNetworkForLogger();
        }
    }

    client->release();
    return loggerList;
}


vector<TSLClusterImpl>
BPNGLoggerDetector::getTSLs(const vector<OnlineLoggerInfoWrapper> &loggersInNetwork)
{
    map<int, TSLClusterImpl> tsls;

    for (const auto &device : loggersInNetwork)
    {
        if (device.getTslNetworkId() != -1)
        {
            if (tsls.find(device.getTslNetworkId()) == tsls.end())
                tsls.emplace(device.getTslNetworkId(), TSLClusterImpl(device.getInfo()));
            else
                tsls[device.getTslNetworkId()].addDevice(device.getInfo());
        }
    }

    vector<TSLClusterImpl> retVal;

    for (const auto &tsl : tsls)
        retVal.push_back(tsl.second);

    return retVal;
}

void BPNGLoggerDetector::excludeRCTFromTSL(TSLClusterImpl &tslChain)
{
    vector<size_t> indices;
    TSLCluster tsl = tslChain.getTSLCluster();

    for (size_t i = 0; i < tsl.numDevices; i++)
    {
        if (tsl.loggerArray[i].deviceType == 4)
            indices.push_back(i);
    }

    for (const auto &index : indices)
        tslChain.deleteDevice(index);
}

int BPNGLoggerDetector::onTargetPathTooLong(char *newTarget, int maxSize)
{
    return 0;

}

int BPNGLoggerDetector::getOverwritingPermission(const char *filePath)
{
    return -1;

}

void BPNGLoggerDetector::onFirmwareUpdateProgress(int percentage, int step, int substep, const char *desc)
{
    if (percentage != oldpercent)
    {
        cout << "LISTENER: FWUpdate: " << setw(3) << setfill(' ') << percentage << "%  " << step << " " << substep << ": " << desc << endl;
        oldpercent = percentage;
    }
}

void BPNGLoggerDetector::onFirmwareUpdateError(int errorId)
{
}

int BPNGLoggerDetector::onCriticalDiskSpace(uint64_t freeSpace, uint64_t neededSpace, const char *drive, const char *msg)
{
    cout << "LISTENER: criticalDiskSpace: free " << freeSpace << " needed " << neededSpace << " on drive " << drive << " comment " << msg << endl;
    return 0;
}

void BPNGLoggerDetector::onLoggerConnected(bool isConnected)
{
}

int BPNGLoggerDetector::onGetLogReportProgress(int percentage, const char *desc)
{
    if (percentage != oldpercent)
    {
        cout << "LISTENER: Conversion: " << setw(3) << setfill(' ') << percentage << "%  " << desc << endl;
        oldpercent = percentage;
    }
    return 1;
}

void BPNGLoggerDetector::onWarning(BPNGWarningCode warningCode, const char *warnMsg)
{
    cout << "LISTENER: warning received " << warningCode << " " << warnMsg << endl;
}


void BPNGLoggerDetector::onInvalidPwConfigFound(const char *)
{
}

void BPNGLoggerDetector::onDownloadStart(int64_t totalAmountOfBytes)
{
    cout << "LISTENER: onDownloadStart: total " << totalAmountOfBytes << " bytes" << endl;
    oldpercent = -1;
}

void BPNGLoggerDetector::onLogInDataFailed()
{
}

void BPNGLoggerDetector::onResetLogInDataFailed()
{
}

void BPNGLoggerDetector::onFuncAccessDenied()
{
}

void BPNGLoggerDetector::onConversionStart(int64_t totalAmountOfBytes)
{
    oldpercent = -1;
}

void
BPNGLoggerDetector::logFunc(const char *s)
{
    cerr << s << endl;
}
