//************************************************************
//
/** @file OnlineLoggerInfoWrapper.hh
 *
 * @brief C++ wrapper around brain dead OnlineLoggerInfo.
 */
//************************************************************

#ifndef ONLINELOGGERINFOWRAPPER_HH
#define ONLINELOGGERINFOWRAPPER_HH

// sys
#include <cstdlib>
#include <cstring>
#include <vector>
#include <iosfwd>

// client
#include <IBPNGClient.h>


//************************************************************
//
// OnlineLoggerInfoWrapper
//
/// Wrapper around brain dead OnlineLoggerInfo.
//
//************************************************************

class OnlineLoggerInfoWrapper : private OnlineLoggerInfo
{
public:

    /// Default CTOR.
    OnlineLoggerInfoWrapper();

    /// CTOR taking an IP and a name.
    OnlineLoggerInfoWrapper(const char *_ip, const char *_name);

    /// CTOR from the "original".
    OnlineLoggerInfoWrapper(const OnlineLoggerInfo &info);

    /// Copy CTOR.
    OnlineLoggerInfoWrapper(const OnlineLoggerInfoWrapper &info);

    /// Move CTOR.
    OnlineLoggerInfoWrapper(OnlineLoggerInfoWrapper &&info);

    /// Copy assignement from the "original".
    OnlineLoggerInfoWrapper & operator= (const OnlineLoggerInfo &info);

    /// Copy assignement.
    OnlineLoggerInfoWrapper & operator= (const OnlineLoggerInfoWrapper &info);

    /// Move assignement.
    OnlineLoggerInfoWrapper & operator= (OnlineLoggerInfoWrapper &&info);

    /// DTOR.
    ~OnlineLoggerInfoWrapper();

    /// Get OnlineLoggerInfo.
    const OnlineLoggerInfo & getInfo() const;

    /// Set logger's IP.
    void setIP(const char *_ip);

    /// Get logger's IP.
    const char * getIP() const;

    /// Set logger's name.
    void setName(const char *_ip);

    /// Get logger's name.
    const char * getName() const;

    /// Set logger's mainboard number.
    void setMbnr(const char *_ip);

    /// Get logger's mainboard number.
    const char * getMbnr() const;

    /// Get logger's TSL network ID.
    int32_t getTslNetworkId() const;

    /// Output operator.
    friend std::ostream & operator<<(std::ostream &os, const OnlineLoggerInfoWrapper &oli);


private:

    /// Return an empty C string on the heap.
    static char *emptyString();

    /// Return a duplicate of @a sorig on the heap.
    static char *dupString(const char *sorig);

    /// Empty all string members.
    static void emptyAll(OnlineLoggerInfoWrapper &dst);

    /// Duplicate all string members.
    static void duplicateAll(OnlineLoggerInfoWrapper &dst, const OnlineLoggerInfo &info);

    /// Delete all string members.
    void deleteAll();
};


// inline and template functions

inline
OnlineLoggerInfoWrapper::OnlineLoggerInfoWrapper()
{
    OnlineLoggerInfo::operator=(createEmptyOnlineLoggerInfo());
    emptyAll(*this);
}


inline
OnlineLoggerInfoWrapper::OnlineLoggerInfoWrapper(const OnlineLoggerInfo &info)
{
    OnlineLoggerInfo::operator=(info);
    duplicateAll(*this, info);
}


inline
OnlineLoggerInfoWrapper::OnlineLoggerInfoWrapper(const OnlineLoggerInfoWrapper &info)
{
    OnlineLoggerInfo::operator=(info);
    duplicateAll(*this, info);
}


inline
OnlineLoggerInfoWrapper::OnlineLoggerInfoWrapper(OnlineLoggerInfoWrapper &&info)
{
    // get all members 1:1
    OnlineLoggerInfo::operator=(info);

    // fill the RHS with empty strings
    emptyAll(info);
}


inline
OnlineLoggerInfoWrapper::~OnlineLoggerInfoWrapper()
{
    deleteAll();
}


inline
OnlineLoggerInfoWrapper &
OnlineLoggerInfoWrapper::operator= (const OnlineLoggerInfo &info)
{
    deleteAll();
    OnlineLoggerInfo::operator=(info);
    duplicateAll(*this, info);

    return *this;
}


inline const OnlineLoggerInfo &
OnlineLoggerInfoWrapper::getInfo() const
{
    return *this;
}


inline const char *
OnlineLoggerInfoWrapper::getIP() const
{
    return ip;
}


inline const char *
OnlineLoggerInfoWrapper::getName() const
{
    return name;
}


inline const char *
OnlineLoggerInfoWrapper::getMbnr() const
{
    return mbnr;
}


inline int32_t
OnlineLoggerInfoWrapper::getTslNetworkId() const
{
    return tslNetworkId;
}


inline OnlineLoggerInfoWrapper &
OnlineLoggerInfoWrapper::operator= (const OnlineLoggerInfoWrapper &info)
{
    return operator= (info.getInfo());
}


extern std::ostream & operator<<(std::ostream &os, const OnlineLoggerInfoWrapper &oli);
extern std::ostream & operator<<(std::ostream &os, const std::vector<OnlineLoggerInfoWrapper> &list);

#endif	// ONLINELOGGERINFOWRAPPER_HH
